@@ -4,7 +4,7 @@
 #
 # timsieved client script
 #
-# Copyright © 2006, 2007 Phil Pennock.  All rights reserved.
+# Copyright © 2006-2008 Phil Pennock.  All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions
@@ -35,7 +35,7 @@ use strict;
 
 my %ssl_options = (
 	SSL_version	=> 'TLSv1',
-	SSL_cipher_list	=> 'ALL:!NULL:!LOW:!EXP:!ADH:@STRENGTH',
+	SSL_cipher_list	=> 'ALL:!aNULL:!NULL:!LOW:!EXP:!ADH:@STRENGTH',
 	SSL_verify_mode	=> 0x01,
 	SSL_ca_path	=> '/etc/ssl/certs',
 );
@@ -48,7 +48,7 @@ my @cmd_localfs_ls = qw( ls -C );
 # No user-serviceable parts below
 
 our $VERSION = 0;
-my $svn_revision = '$Revision: 41 $';
+my $svn_revision = '$Revision: 57 $';
 if ($svn_revision =~ /^.Revision:\s*(\d+)\s*\$\z/) {
 	$svn_revision = $1;
 	$VERSION = '0.' . $1;
@@ -56,10 +56,18 @@ if ($svn_revision =~ /^.Revision:\s*(\d+)\s*\$\z/) {
 	$svn_revision = '0 because unknown';
 }
 
-use Authen::SASL qw(Perl); # Need a way to ask which mechanism to send
+use Authen::SASL 2.11 qw(Perl);
+# 2.11: first version with non-broken DIGEST-MD5
+#       Earlier versions don't allow server verification
+#       NB: code still explicitly checks for a new-enough version, so
+#           if you have an older version of Authen::SASL and know what you're
+#           doing then you can remove this version check here.  I advise
+#           against it, though.
+# Perl: Need a way to ask which mechanism to send
 use Authen::SASL::Perl::EXTERNAL; # We munge inside its private stuff.
 use Cwd qw();
 use Errno;
+use File::Basename qw();
 use Getopt::Long;
 use IO::File;
 use IO::Socket::INET6;
@@ -86,6 +94,7 @@ my $localsievename;
 my $remotesievename;
 my ($user, $authzid, $authmech, $sslkeyfile, $sslcertfile, $passwordfd);
 my $prioritise_auth_external = 0;
+my $dump_tls_information = 0;
 my ($server, $realm);
 my $port = 'sieve(2000)';
 my $net_domain = AF_UNSPEC;
@@ -109,6 +118,7 @@ GetOptions(
 	"4"		=> sub { $net_domain = AF_INET },
 	"6"		=> sub { $net_domain = AF_INET6 },
 	"debug"		=> \$DEBUGGING,
+	"dumptlsinfo|dumpsslinfo"	=> \$dump_tls_information,
 	# option names can be short-circuited, $action is complete:
 	"upload"	=> sub { $action = 'upload' },
 	"download"	=> sub { $action = 'download' },
@@ -212,6 +222,7 @@ my $sock = IO::Socket::INET6->new(
 	PeerPort	=> $port,
 	Proto		=> 'tcp',
 	Domain		=> $net_domain,
+	MultiHomed	=> 1, # try multiple IPs (IPv4 works, v6 doesn't?)
 );
 unless (defined $sock) {
 	my $extra = '';
@@ -227,21 +238,59 @@ debug "connection: remote host address is @{[$sock->peerhost()]}";
 my %capa;
 my %raw_capabilities;
 my %capa_dosplit = map {$_ => 1} qw( SASL SIEVE );
+# Key is permissably empty keyword, value if defined is closure to call with
+# capabilities after receiving complete list, for verifying permissability.
+# First param $sock, second \%capa, third \%raw_capabilities
+my %capa_permit_empty = (
+	# draft 7 onwards clarify that empty SASL is permitted, but is error
+	# in absense of STARTTLS
+	SASL	=> sub {
+		return if exists $_[1]{STARTTLS};
+		# We die because there's no way to authenticate.
+		# Spec states that after STARTTLS SASL must be non-empty
+		closedie $_[0], "Empty SASL not permitted without STARTTLS\n";
+		},
+	SIEVE	=> undef,
+);
 
 sub parse_capabilities
 {
 	my $sock = shift;
-	my $external_first = shift;
-	$external_first = 0 unless defined $external_first;
+	local %_ = @_;
+	my $external_first = 0;
+	$external_first = $_{external_first} if exists $_{external_first};
 
+	my @double_checks;
 	%raw_capabilities = ();
 	%capa = ();
 	while (<$sock>) {
-		chomp; s/\s*$//;
 		received;
-		last if /^OK$/;
-		if (/^\"([^"]+)\"\s+\"(.+)\"$/) {
-			my ($k, $v) = ($1, $2);
+		chomp; s/\s*$//;
+		if (/^OK\b/) {
+			last unless exists $_{until_see_no};
+		} elsif (/^\"([^"]+)\"\s+\"(.*)\"$/) {
+			my ($k, $v) = (uc($1), $2);
+			unless (length $v) {
+				unless (exists $capa_permit_empty{$k}) {
+					warn "Empty \"$k\" capability spec not permitted: $_\n";
+					# Don't keep the advertised capability unless
+					# it has some value which is needed.  Eg,
+					# NOTIFY must list a mechanism to be useful.
+					next;
+				}
+				if (defined $capa_permit_empty{$k}) {
+					push @double_checks, $capa_permit_empty{$k};
+				}
+			}
+			if (exists $capa{$k}) {
+				# won't catch if the first instance was ignored for an
+				# impermissably empty value; by this point though we
+				# would already have issued a warning and the server
+				# is so fubar that it's not worth worrying about.
+				warn "Protocol violation.  Already seen capability \"$k\".\n" .
+					"Ignoring second instance and continuing.\n";
+				next;
+			}
 			$raw_capabilities{$k} = $v;
 			$capa{$k} = $v;
 			if (exists $capa_dosplit{$k}) {
@@ -250,10 +299,34 @@ sub parse_capabilities
 		} elsif (/^\"([^"]+)\"$/) {
 			$raw_capabilities{$1} = '';
 			$capa{$1} = 1;
+		} elsif (/^NO\b/) {
+			return if exists $_{until_see_no};
+			warn "Unhandled server line: $_\n"
+		} elsif (/^BYE\b(.*)/) {
+			closedie_NOmsg $sock, $1,
+				"Server said BYE when we expected capabilities.\n";
 		} else {
 			warn "Unhandled server line: $_\n"
 		}
+	};
+
+	closedie $sock, "Server does not return SIEVE capability, unable to continue.\n"
+		unless exists $capa{SIEVE};
+	warn "Server does not return IMPLEMENTATION capability.\n"
+		unless exists $capa{IMPLEMENTATION};
+
+	foreach my $check_sub (@double_checks) {
+		$check_sub->($sock, \%capa, \%raw_capabilities);
 	}
+
+	if (grep {lc($_) eq 'enotify'} @{$capa{SIEVE}}) {
+		unless (exists $capa{NOTIFY}) {
+			warn "enotify extension present, NOTIFY capability missing\n" .
+				"This violates MANAGESIEVE specification.\n" .
+				"Continuing anyway.\n";
+		}
+	}
+
 	if (exists $capa{SASL} and $external_first
 			and grep {uc($_) eq 'EXTERNAL'} @{$capa{SASL}}) {
 		# We do two things.  We shift the EXTERNAL to the head of the
@@ -285,9 +358,35 @@ if (exists $capa{STARTTLS}) {
 		die "STARTTLS promotion failed: $e\n";
 	};
 	debug("--- TLS activated here");
+	if ($dump_tls_information) {
+		print $sock->dump_peer_certificate();
+		if ($DEBUGGING and
+		    exists $main::{"Net::"} and exists $main::{"Net::"}{"SSLeay::"}) {
+			# IO::Socket::SSL depends upon Net::SSLeay
+			# so this should be fairly safe, albeit messing
+			# around behind IO::Socket::SSL's back.
+			print STDERR Net::SSLeay::PEM_get_string_X509(
+				$sock->peer_certificate());
+		}
+	}
 	$forbid_clearauth = 0;
-	ssend $sock, "CAPABILITY";
-	parse_capabilities($sock, $prioritise_auth_external);
+	# Cyrus sieve might send CAPABILITY after STARTTLS without being
+	# prompted for it.  This breaks the command-response model.
+	# We can't just check to see if there's data to read or not, since
+	# that will break if the next data is delayed (race condition).
+	# There is no protocol-compliant method to determine this, short
+	# of "wait a while, see if anything comes along; if not, send
+	# CAPABILITY ourselves".  So, I break protocol by sending the
+	# non-existent command NOOP, then scan for the resulting NO.
+	ssend $sock, "NOOP";
+	parse_capabilities($sock,
+		until_see_no	=> 1,
+		external_first => $prioritise_auth_external);
+	unless (scalar keys %capa) {
+		ssend $sock, "CAPABILITY";
+		parse_capabilities($sock,
+			external_first => $prioritise_auth_external);
+	}
 } elsif ($forbid_clearchan) {
 	die "TLS not offered, SASL confidentiality not supported in client.\n";
 }
@@ -415,13 +514,24 @@ if (defined $realm) {
 		# so if it says "okay", we don't keep trying.
 		my $final_auth = decode_base64($1);
 		my $valid = $authconversation->client_step($final_auth);
-		# Skip checking $authconversation->code() here because
+		# With Authen::SASL before 2.11 (..::Perl 1.06),
 		# Authen::SASL::Perl::DIGEST-MD5 module will complain at this
 		# final step:
 		#   Server did not provide required field(s): algorithm nonce
 		# which is bogus -- it's not required or expected.
+		# Authen::SASL 2.11 fixes this, with ..::Perl 1.06
+		# We explicitly permit silent failure with the security
+		# implications because we require a new enough version of
+		# Authen::SASL at import time above and if someone removes
+		# that check, then on their head be it.
+		if ($authconversation->code()) {
+			my $emsg = $authconversation->error();
+			if ($Authen::SASL::Perl::VERSION >= 1.06) {
+				closedie($sock, "SASL Error: $emsg\n");
+			}
+		}
 		if (defined $valid and length $valid) {
-			closedie($sock, "Server failed final verification");
+			closedie($sock, "Server failed final verification [$valid]");
 		}
 	}
 
@@ -626,10 +736,13 @@ if (defined $execscript) {
 		closedie $sock, "No terminal initialisation"
 			unless defined $term;
 		$term->ornaments(0);
-		$term->Attribs->{completion_function} =
-			sub { complete_rl_sieve($term, $sock, @_) };
-		$term->Attribs->{completer_quote_characters} = '"';
-		$term->Attribs->{filename_quote_characters} = " \t";
+		if ($term->ReadLine() =~ /::Gnu/) {
+			# The relevant hooks aren't in the Perl implementation
+			$term->Attribs->{completion_function} =
+				sub { complete_rl_sieve($term, $sock, @_) };
+			$term->Attribs->{completer_quote_characters} = '"';
+			$term->Attribs->{filename_quote_characters} = " \t";
+		}
 		$cmdlineget_func = sub { return $term->readline('> ') };
 		print STDERR "ReadLine support enabled.\n";
 	};
@@ -717,6 +830,11 @@ while (defined (my $cmdline = $cmdlineget_func->())) {
 	}
 	if ($minp != -1 and (@params != $maxp and @params)) {
 		my $repeat = $params[-1];
+		# When repeating it's assumed to be a filename.  There may
+		# be an issue with putting a file from a different directory,
+		# should take basename for repeats.  I can't think of a
+		# situation where basename wouldn't be correct.
+		$repeat = File::Basename::basename($repeat);
 		for (my $i = $#params+1; $i <= $maxp; ++$i) {
 			$params[$i] = $repeat;
 		}
@@ -839,7 +957,7 @@ sub sieve_download
 		warn qq{Empty script "$remotefn"?  Not saved.\n};
 		return;
 	}
-	unless (/^{(\d+)}\r?$/m) {
+	unless (/^{(\d+)\+?}\r?$/m) {
 		die "QUIT:Failed to parse server response to GETSCRIPT";
 	}
 	my $contentdata = $_;
@@ -1079,8 +1197,14 @@ sub debug
 	print STDERR "$_[0]\n";
 }
 
-sub sent { $_[0] = $_ unless defined $_[0]; debug ">>> $_[0]"; }
-sub received { $_[0] = $_ unless defined $_[0]; debug "<<< $_[0]"; }
+sub diag {
+	my ($prefix, $data) = @_;
+	$data =~ s/\r/\\r/g; $data =~ s/\n/\\n/g; $data =~ s/\t/\\t/g;
+	$data =~ s/([^[:graph:] ])/sprintf("%%%02X", ord $1)/eg;
+	debug "$prefix $data";
+}
+sub sent { my $t = defined $_[0] ? $_[0] : $_; diag('>>>', $t) }
+sub received { my $t = defined $_[0] ? $_[0] : $_; diag('<<<', $t) }
 
 sub ssend
 {
@@ -1095,7 +1219,7 @@ sub ssend
 # yes, the debug output can have extra blank lines if supplied -noeol because
 # they're already present.  Rather than mess around to tidy it up, I'm leaving
 # it because it's debug output, not UI or protocol text.
-		sent $l;
+		sent "$l$eol";
 	}
 }
 
@@ -1106,6 +1230,10 @@ sub sget
 	$dochomp = 0 if defined $_[0] and $_[0] eq '-nochomp';
 	my $l;
 	$l = $sock->getline();
+	unless (defined $l) {
+		debug "... no line read, connection dropped?";
+		die "Connection dropped unexpectedly when trying to read.\n";
+	}
 	if ($l =~ /{(\d+)\+?}\s*\n?\z/) {
 		debug "... literal string response, length $1";
 		my $len = $1;
@@ -1120,10 +1248,10 @@ sub sget
 		}
 		$dochomp = 0;
 	}
+	received $l;
 	if ($dochomp) {
 		chomp $l; $l =~ s/\s*$//;
 	}
-	received $l;
 	if (defined wantarray) {
 		return $l;
 	} else {
@@ -1219,7 +1347,7 @@ sieve-connect - managesieve command-line client
                [-m <authmech>] [-r realm] [-e execscript]
 	       [... longopts ...]
  sieve-connect [--localsieve <script>] [--remotesieve <script>]
-	       [--debug]
+	       [--debug] [--dumptlsinfo]
                [--server <hostname>] [--port <portspec>] [--4|--6]
 	       [--user <authentication_id>] [--authzid <authzid>]
 	       [--realm <realm>] [--passwordfd <n>]
@@ -1267,6 +1395,9 @@ people think about names in the local filesystem.  There is no default
 script name.
 
 The B<--debug> option turns on diagnostic traces.
+The B<--dumptlsinfo> shows the TLS (SSL) peer information; if specified
+together with B<--debug> then the server's PEM certificate will be
+provided as debug trace.
 
 The server can be a host or IP address, IPv4 or IPv6;
 the default is C<$IMAP_SERVER> from the environment (if it's not a
